/*
	Copyright (C) 2000 Davide Bianchi
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You can found the complete GNU GPL at:
	http://www.fsf.org/copyleft/gpl.html
	
	Or write to the Free Software Foundation,
	Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	
	Davide Bianchi
	
	davide_bianchi@usa.net
	Kapellerhof 82
	6045DD Roermond
	The Netherlands

*/
import java.net.Socket;
import java.net.UnknownHostException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.FileOutputStream;
import java.util.List;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.sql.*;

/*
	Build the Index for the search engine
*/	
public class IBuild
{

	private Socket sock;
	
	private InputStream in;
	private OutputStream out;
	
	private Connection conn;
	private PreparedStatement st;
	private PreparedStatement st1;
	private ResultSet rs;

	private FileOutputStream outFile;
	private FileOutputStream index;
	private List indexList;
	
	private String savedata;
	private String currentUrl;
	private String host;
	private String title;
	private List urls;

	private int current;
	private int totUrl;
	private int totWord;

	public static final int NOTAG = 0;
	public static final int TITLE = 1;
	public static final int HREF = 2;

	public static final int MINLEN = 3;

	private boolean inBody=false;
	private boolean inHead=false;
	private boolean preHtml=true;
	private boolean inTitle=false;

	private boolean testOnly = false;
	
	private final String extStrip[] = { "www.",
                                      "http://", 
                                      "ftp://", 
                                      "file://",
                                      "mailto:", 
                                      ".pdf",
																		  ".png",
                                      ".jpg",
                                      ".gif",
                                      ".doc",
                                      ".rtf",
                                      ".zip",
                                      ".tar.gz",
                                      ".tar",
                                      ".ps" };
	private int tagId=0;

	/* log info */
	private void log( String text )
	{
		if( outFile != null )
		{
			try
			{
				text += "\n";
				outFile.write( text.getBytes() );
			}
			catch( IOException e )
			{
				System.out.println( text );
			}
		}
	}

  /* check if the current link is already in the list */
  public boolean alreadyFound( String data )
	{
		boolean result = false;

		String value = title + "@" + currentUrl + "@" + data + "\n";
		if( indexList.contains( value ))
		{
			result = true;
		}

		return result;

	}

	/* set test function only */
	public void setTest()
	{
		testOnly = true;
	}

	/* check a Tag */
	private void checkTag( String tag )
	{
		String subs;

		/* put the tag to lower case to avoid problem */
		tag = tag.toLowerCase();

		/* begin html */		
		if( tag.compareTo("<html") == 0 )
			preHtml = false;
		
		/* begin of title */
		if( tag.compareTo("<title") == 0 )
			tagId = TITLE;

		/* end of title */
		if( tag.compareTo("</title") == 0 )
			tagId = NOTAG;

		/* begin of the body */
		if( tag.compareTo("<body") == 0 )
			inBody = true;

		/* begin script section */
		if( tag.compareTo("<script") == 0 )
			inBody = true;

		/* end of script section */
		if( tag.compareTo("</script") == 0 )
			inBody = true;

		/* end of the body */
		if( tag.compareTo("</body") == 0 )
			inBody = false;

		/* check for HREF */
		if( tag.indexOf("href") >= 0  )
		{
			/* parse the HREF */
			tagId = HREF;
			parseLink( tag );
		}
	}

	/* parse an hyperlink */
	private void parseLink( String link )
	{
		StringBuffer l=new StringBuffer("");
		StringTokenizer t = new StringTokenizer( link, " " );
		String p="";
		String q="";
		int i=0;
		int in=0;

		/* look for the HREF part of the tag */
		while( t.hasMoreTokens() )
		{
			p = t.nextToken();

			if(p.length() > 4 )
			{
				if( p.substring(0,4).compareTo("href") == 0 )
				{
					/* gotcha! */
					break;
				}
			}
		}

		/* loop and get the real URL */
		for( i=0; i < p.length(); i++ )
		{
			if( p.charAt(i) == 34 || p.charAt(i) == 39 )
			{
				in++;
			}

			if( in == 1)
			{
				l.append( p.charAt(i) );
			}
		}

		/* now remove the first char (") from the URL */
		if( l.length() > 1 )
		{
			p = l.toString();
			p = p.substring( 1 );

			/* check if the link is a valid one */
			for( i=0; i < extStrip.length; i++ )
			{
				if( p.indexOf( extStrip[i] ) >= 0 )
				{
					/* wrong! */
					log( "found link " + p + " in " + currentUrl + " - discarded." );
					p = "";
					break;
				}
			}

			/* Ok, add the link to the list */
			if( p.compareTo("") != 0 )
			{
				addUrl( p );
			}
		}
	}

	/* process normal words */
	private void processWords( String words )
	{
		StringTokenizer t = new StringTokenizer( words, "\n\t\" ;.:,'<>/?{}[]()*#@$!+-" );
		String p = "";
		int i;

		/* loop and process each token */
		while( t.hasMoreTokens() )
		{
			/* get the token and convert in lower case*/
			p = t.nextToken();
			p = p.toLowerCase();

			/* remove eventually &xgrave... */
			if( (i = p.indexOf("&agrave")) >= 0 )
			{
				if( i == 0 )
					break;
				p = p.substring(0, i-1 ) + "a";
			}
			
			if( (i = p.indexOf("&egrave")) >= 0 )
			{
				if( i == 0 )
					break;
				p = p.substring(0, i-1 ) + "e";
			}

			if( (i = p.indexOf("&igrave")) >= 0 )
			{
				if( i == 0 )
					break;
				p = p.substring(0, i-1 ) + "i";
			}

			if( (i = p.indexOf("&ograve")) >= 0 )
			{
				if( i == 0 )
					break;
				p = p.substring(0, i-1 ) + "o";
			}

			if( (i = p.indexOf("&ugrave")) >= 0 )
			{
				if( i == 0 )
					break;
				p = p.substring(0, i-1 ) + "u";
			}

			/* process only word more than "MINLEN" */
			if( p.length() >= MINLEN )
			{
				/* save the data */
				writeData( p );
			}
		}
	}

	/* parse the buffer */
	private void parseBuffer( String buffer )
	{
		/* loop and manage the various element */
		int i=0;
		String tag="";
		String word="";
		StringBuffer x= new StringBuffer("");
		boolean process = false;

		char el;
		boolean inTag=false;
		inBody = false;
		preHtml = true;

		/* loop and read each element */
		for( i=0; i < buffer.length(); i++ )
		{
			/* read the element */
			process=true;
			el = buffer.charAt(i);

			/* remove tabs, new line etc. */
			if( el == '\t' || el == '\n' || el < 32 )
			{
				el = ' ';
			}

			/* check begin or end of a tag */
			if( el == '<' )
			{
				/* If I was in the Title, collect the title */
				if( inTitle )
				{
					/* remove trailing spaces */
					title = x.toString().trim();
					inTitle = false;
				}
				else
				{
					/* If I was in the body, process the text */
					if( inBody && word.length() > 0 )
					{
						processWords( word );
					}
				}

				preHtml= false;
				inTag = true;
				x = new StringBuffer();
			}

			if( el == '>' )
			{
				tag = x.toString();
				checkTag( tag );

				/* check what tag I found */
				switch( tagId )
				{
					case IBuild.TITLE:
						inTitle = true;
						break;
						
					case IBuild.HREF:
						break;
				}

				process=false;
				inTag = false;
				inBody = true;
				x = new StringBuffer("");

			}

			/* if valid char, add to the temporary buffer */
			if( process )
				x.append( el );

			if( inBody )
			{
				/* I'm in the body of the page, get the word */
				word = x.toString();
			}

			if( inTitle )
			{
				/* I'm in the title */
				title = x.toString();
			}
		}
	}

	/* return the current url removing the trailing document */
	private String getCurrentUrl()
	{
		String url = currentUrl;
		int i;

		/* remove trailing doc */
		/* loop backward and stop at the first "/" */
		for( i=url.length()-1; i>=0; i-- )
		{
			/* check if char is a "/" */
			if( url.charAt( i ) == '/' )
			{
				url = url.substring( 0, i+1 );
				return url;
			}
		}

		/* if I'm here, means that the url is on the root dir */
		return "/";

	}

	/* add an Url to the list
		 check also if the URL is absolute (eg: begin with "/") or
		 relative, in this case I build an absolute URL
	*/
	public void addUrl( String url )
	{
		int i=0;

		/* check if the url is relative to this document */
		if( url.startsWith("#") )
		{
			/* does nothing */
			return;
		}
		
		/* check if the URL contain a relative position into a document (#) */
		if( url.indexOf('#') > 0 )
		{
			/* remove the relative position */
			i = url.indexOf('#');
			url = url.substring( 0, i );
		}
		
		/* check if the url is absolute or not */
		if( ! url.startsWith("/") )
		{
			/* url is relative, add the current Url */
			url = getCurrentUrl( ) + url;
		}

		if( ! urls.contains( url ) )
		{
			totUrl++;
			urls.add( url );
		}
	}

	/* perform scanning of an entire host */
	public void scan( )
	{
		String url="";

		/* check */
		if( host == null )
		{
			return;
		}

		if( urls.size() == 0 )
		{
			addUrl( "/" );
		}

		try
		{
			/* open the database and log file connection */
			if( ! openOutput() )
				return;

			while( true )
			{
				/* check if I have exausted the url list */
				if( current < urls.size() )
				{
					/* retrive an element */
					url = (String) urls.get( current );
					current++;

					if( url == null )
					{
						break;
					}

					if( ! url.startsWith("/") )
					{
						url = "/" + url;
					}
				}
				else
				{
					break;
				}

				/* put out a message for scanning */
				System.out.println("Scanning " + url );

				/* scan the URL */
				scanUrl( url );

			}

			/* display results */
			System.out.println("Total urls scanned: " + totUrl );
			System.out.println("Total word scanned: " + totWord );

			/* close all the output file */
			closeOutput( );

		}
		catch( Exception e )
		{
			/* exception are not managed */
			e.printStackTrace();
		}
	}

	/* open the output file and/or database used to store info
	************************************************************
	NOTE: modify this to open the "real" database connection
	************************************************************
	*/
	private boolean	openOutput( )
	{
		try
		{
			outFile = new FileOutputStream( "scan.log" );
			
			if( ! testOnly )
			{
				/* open the log file to write data */
				index = new FileOutputStream("web.index");
			}
		}
		catch( Exception e )
		{
			System.out.println( e.getMessage() );
			return false;
		}

		return true;
	}

	/* close the file or database used for output */
	private void closeOutput( )
	{
		try
		{
			outFile.close();

			if( ! testOnly )
			{
				index.close();
			}
		}
		catch( Exception e )
		{
		}
	}

	/* insert the data into the file/database */
	private void writeData( String data )
	{
		String text;

		/* do not insert titleless document */
		if( title.compareTo("") == 0 )
			return;

		/* if is only a test do not write data for real */
		if( testOnly )
		{
			/* increase total word counter */
			totWord++;
			return;
		}

		/* write data */
		try
		{
			/* build text */
      if( ! alreadyFound( data ) )
			{ 
			   text = title + "@" + currentUrl + "@" + data + "\n";
         index.write( text.getBytes() );
				 indexList.add( text );

			   /* increase total word counter */
			   totWord++;
			}
		}
		catch( IOException e )
		{
			/* log errors */
			log( "Error inserting data in the database" + e.getMessage() );
		}
	}

	/* scan an url */
	void scanUrl( String url )
	{

		String request;
		byte b[] = new byte[4096];
		int bytes;
		savedata = "";
		String data="";
		String good="";
		String err;

		/* clear the last title */
		title = "";

		/* build the HTTP request */
		request = "GET " + url + " HTTP/1.0\n\n";

		try
		{
			/* open the socket */
			sock = new Socket( host, 80 );

			/* open the stream and the output file */
			in = sock.getInputStream();
			out = sock.getOutputStream();

			/* send request */
			out.write( request.getBytes() );

			/* set the current url */
			currentUrl = url;

			/* get info */
			while( true )
			{
				/* read a bunch of bytes */
				if( (bytes=in.read( b )) > 0 )
				{
					/* add the data to the buffer */
					data = data + new String( b, 0, bytes );
				}
				else
				{
					/* check result, discard errors */
					/* a "good" response contain HTTP/1.1 200 OK */
					if( data.indexOf( "200 OK" ) >= 0 )
					{
						/* Ok, all data read, now parse the buffer */
						parseBuffer( data );
						break;
					}
					else
					{
						/* error retriving data, log */
						log("Error retriving data for " + currentUrl);
						break;
					}
				}
			}

			/* close the socket */
			sock.close();

		}
		catch( UnknownHostException e )
		{
			System.out.println("Host " + host + " not found.");
			System.exit(0);
		}
		catch( IOException e )
		{
			System.out.println("Error during socket creation.");
			System.exit(0);
		}
	}

	/* set the host */
	public void setHost( String host )
	{
		if( host.startsWith("http:") )
		{
			host = host.substring(7);
		}

		this.host = host;
	}

	/* constructor */
	public IBuild()
	{
		savedata = "";
		totUrl = 0;
		totWord = 0;
		host = null;
		urls = new ArrayList();
		indexList = new ArrayList();
		current=0;
	}

	/* constructor with initialization */
	public IBuild( String host )
	{
		/* clear totals */
		savedata = "";
		totUrl = 0;
		totWord = 0;

		/* build host name */
		setHost( host );

		/* build url list */
		indexList = new ArrayList();
		urls = new ArrayList();
		current = 0;

		/* build the current url */
		addUrl( "/" );
	}

	/* constructor with initialization */
	public IBuild( String host, String url )
	{
		/* clear totals */
		savedata = "";
		totUrl = 0;
		totWord = 0;

		/* build host name */
		setHost( host );

		/* build url list */
		indexList = new ArrayList();
		urls = new ArrayList();
		current = 0;

		/* build the current url */
		addUrl( url );
	}

	/* main program */
	public static void main( String args[] )
	{
		/* check parameter */
		IBuild.showTitle( 0 );

		if( args.length < 1 )
		{
			System.out.println("Usage: java IBuild -h host [-u URL] [-t] [-w]");
		}
		else
		{
			IBuild i = new IBuild();

			/* check command line options */
			for( int j=0; j < args.length; j++ )
			{
				/* host */
				if( args[j].compareTo( "-h" )== 0 || args[j].compareTo( "-H" )== 0 )
				{
					i.setHost( args[j+1] );
				}

				/* specific url */
				if( args[j].compareTo( "-u" )== 0 || args[j].compareTo( "-U" )== 0 )
				{
					i.addUrl( args[j+1] );
				}

				/* show warning and exit */
				if( args[j].compareTo( "-w" )== 0 || args[j].compareTo( "-W" )== 0 )
				{
					showTitle(1);
				}

				/* test only */
				if( args[j].compareTo( "-t" )== 0 || args[j].compareTo( "-T" )== 0 )
				{
					i.setTest();
				}
			}

			/* scan */
			i.scan();
			
		}

		/* end of application */
		System.exit(0);

	}

	/* display software license and other stuff */
	public static void showTitle( int mode )
	{

		System.out.println("IBuild v. 1.00 - Copyright (C) 2000 Davide Bianchi");

		if( mode == 0 )
		{
			System.out.println("IBuild comes with ABSOLUTELY NO WARRANTY; for details type IBuild -w");
			System.out.println("See the GNU License at http://www.fsf.org/copyleft/gpl.html");
			System.out.println("");
		}

		if( mode == 1 )
		{
			System.out.println("");
			System.out.println("BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY FOR");
			System.out.println("THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN");
			System.out.println("OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES");
			System.out.println("PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED");
			System.out.println("OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF");
			System.out.println("MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK AS");
			System.out.println("TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU. SHOULD THE");
			System.out.println("PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,");
			System.out.println("REPAIR OR CORRECTION. ");
			System.out.println("");
			System.out.println("IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING");
			System.out.println("WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR");
			System.out.println("REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR");
			System.out.println("DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL");
			System.out.println("DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM");
			System.out.println("(INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED");
			System.out.println("INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE");
			System.out.println("OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER");
			System.out.println("OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. ");

		}
	}
}
